import {getAssets} from "./assets.ts";
import fontSheet from '../../art/fonts/vga_8x16.png';
import {AlignX, AlignY, Color, Point, Size} from "../datatypes.ts";

class Font {
  #filename: string;
  #cellsPerSheet: Size;
  #pixelsPerCell: Size;
  #tintingCanvas: HTMLCanvasElement;
  #tintedVersions: Record<string, HTMLImageElement>;

  constructor(filename: string, cellsPerSheet: Size, pixelsPerCell: Size) {
    this.#filename = filename;
    this.#cellsPerSheet = cellsPerSheet;
    this.#pixelsPerCell = pixelsPerCell;
    this.#tintingCanvas = document.createElement("canvas");
    this.#tintedVersions = {}
  }

  get #cx(): number { return this.#cellsPerSheet.w }
  get #cy(): number { return this.#cellsPerSheet.h }
  get #px(): number { return this.#pixelsPerCell.w }
  get #py(): number { return this.#pixelsPerCell.h }

  #getTintedImage(color: string): HTMLImageElement | null {
    let image = getAssets().getImage(this.#filename);

    if (!image.complete) { return null; }

    let tintedVersion = this.#tintedVersions[color];
    if (tintedVersion != undefined) {
      return tintedVersion;
    }

    let w = image.width;
    let h = image.height;

    if (!(w == this.#cx * this.#px && h == this.#cy * this.#py)) {
      throw `unexpected image dimensions for font ${this.#filename}: ${w} x ${h}`
    }

    this.#tintingCanvas.width = w;
    this.#tintingCanvas.height = h;
    let ctx = this.#tintingCanvas.getContext("2d")!;
    ctx.clearRect(0, 0, w, h);
    ctx.drawImage(image, 0, 0);
    ctx.globalCompositeOperation = "source-in";
    ctx.fillStyle = color;
    ctx.rect(0, 0, w, h);
    ctx.fill();

    let result = new Image();
    result.src = this.#tintingCanvas.toDataURL();
    this.#tintedVersions[color] = result;
    return result;
  }

  internalDrawText({ctx, text, position, alignX, alignY, forceWidth, color}: {
    ctx: CanvasRenderingContext2D,
    text: string,
    position: Point, alignX?: AlignX, alignY?: AlignY,
    forceWidth?: number, color: Color
  }) {
    alignX = alignX == undefined ? AlignX.Left : alignX;
    alignY = alignY == undefined ? AlignY.Top : alignY;
    forceWidth = forceWidth == undefined ? 65535 : forceWidth;

    let image = this.#getTintedImage(color.toStyle())
    if (image == null) {
      return;
    }

    let sz = this.#glyphwise(text, forceWidth, () => {});
    let offsetX = position.x;
    let offsetY = position.y;
    offsetX += (alignX == AlignX.Left ? 0 : alignX == AlignX.Center ? -sz.w / 2 : - sz.w)
    offsetY += (alignY == AlignY.Top ? 0 : alignY == AlignY.Middle ? -sz.h / 2 : - sz.h)

    this.#glyphwise(text, forceWidth, (cx, cy, char) => {
      let srcIx = char.charCodeAt(0);
      this.#drawGlyph({ctx: ctx, image: image, ix: srcIx, x: offsetX + cx * this.#px, y: offsetY + cy * this.#py});
    })
  }

  #drawGlyph({ctx, image, ix, x, y}: {ctx: CanvasRenderingContext2D, image: HTMLImageElement, ix: number, x: number, y: number}) {
    let srcCx = ix % this.#cx;
    let srcCy = Math.floor(ix / this.#cx);
    let srcPx = srcCx * this.#px;
    let srcPy = srcCy * this.#py;
    ctx.drawImage(
      image,
      srcPx, srcPy, this.#px, this.#py,
      Math.floor(x), Math.floor(y), this.#px, this.#py
    );
  }

  measureText({text, forceWidth}: {text: string, forceWidth?: number}): Size {
    return this.#glyphwise(text, forceWidth, () => {});
  }

  #glyphwise(text: string, forceWidth: number | undefined, callback: (x: number, y: number, char: string) => void): {w: number, h: number} {
    let cx = 0;
    let cy = 0;
    let cw = 0;
    let ch = 0;
    let wcx = forceWidth == undefined ? undefined : Math.floor(forceWidth / this.#px);

    for (let i = 0; i < text.length; i++) {
      let char = text[i]
      if (char == '\n') {
        cx = 0;
        cy += 1;
        ch = cy + 1;
      } else {
        // console.log("cx, cy, char", [cx, cy, char])
        if (wcx != undefined && cx >= wcx) {
          cx = 0;
          cy += 1;
          ch = cy + 1;
        }

        callback(cx, cy, char)
        cx += 1;
        cw = Math.max(cw, cx);
        ch = cy + 1;
      }
    }

    return { w: cw * this.#px, h: ch * this.#py };
  }
}

export let mainFont = new Font(fontSheet, new Size(32, 8), new Size(8, 16));