import {Grid, Point, Rect, Size} from "./engine/datatypes.ts";
import {ConceptualCell, maps} from "./maps.ts";
import {ALL_STATS, Stat} from "./datatypes.ts";
import {DrawPile} from "./drawpile.ts";
import {D} from "./engine/public.ts";
import {sprDrips, sprRaccoonWalking, sprStatPickup} from "./sprites.ts";
import {BG_INSET, FG_TEXT} from "./colors.ts";
import {getPlayerProgress} from "./playerprogress.ts";

export type MapCellContent =
  {type: "statPickup", stat: Stat} |
  {type: "stairs"} |
  {type: "empty"} |
  {type: "block"}

export type MapCell = {
  content: MapCellContent,
  isValidSpawn: boolean,
  revealed: boolean,
  nextMoveAccessible: boolean,
}

export class HuntMode {
  depth: number
  cells: Grid<MapCell>
  player: Point

  drawpile: DrawPile
  frame: number

  constructor({depth, cells, player}: {depth: number, cells: Grid<MapCell>, player: Point }) {
    this.depth = depth;
    this.cells = cells;
    this.player = player;

    this.drawpile = new DrawPile();
    this.frame = 0;
  }

  // == map generator ==
  static generate({depth}: {depth: number}) {
    let mapNames: Array<string> = Object.keys(maps);
    let mapName = mapNames[Math.floor(Math.random() * mapNames.length)];
    let map = maps[mapName];

    let baseCells = map.map((ccell, _xy) => {
      return this.#generateCell(ccell);
    })

    let cells = new Grid(
      new Size(baseCells.size.w + 2, baseCells.size.h + 2), (xy) => {
        let offset = xy.offset(new Point(-1, -1));
        if (offset.x == -1 || offset.y == -1 || offset.x == baseCells.size.w || offset.y == baseCells.size.h) {
          return this.#generateBoundaryCell();
        }
        return baseCells.get(offset)
      }
    )

    let validSpawns = [];
    for (let x = 0; x < cells.size.w; x++) {
      for (let y = 0; y < cells.size.h; y++) {
        let position = new Point(x, y);
        if (cells.get(position).isValidSpawn)  {
          validSpawns.push(position);
        }
      }
    }
    let player = choose(validSpawns);
    cells.get(player).content = {type: "empty"};

    if (Math.random() < 0.75) {
      while (true) {
        let x = Math.floor(Math.random() * cells.size.w);
        let y = Math.floor(Math.random() * cells.size.h);
        let xy = new Point(x, y);

        let item = cells.get(new Point(x, y));
        if (player.equals(xy)) {
          continue;
        }
        if (item.content.type == "block") {
          continue;
        }
        item.content = {type: "stairs"}
        break;
      }
    }

    let hm = new HuntMode({depth, cells, player})
    hm.#updateVisibilityAndPossibleMoves();
    return hm;
  }

  static #generateCell(conceptual: ConceptualCell): MapCell {
    switch (conceptual) {
      case "X":
        return { content: {type: "block"}, revealed: false, isValidSpawn: false, nextMoveAccessible: false};
      case " ":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: false, nextMoveAccessible: false };
      case ".":
        return { content: HuntMode.#generateContent(), revealed: false, isValidSpawn: true, nextMoveAccessible: false };
    }
  }

  static #generateBoundaryCell() {
    return this.#generateCell("X");
  }

  static #generateContent(): MapCellContent {
    // stat pickup
    let gsp = (): MapCellContent => {
      return {type: "statPickup", stat: choose(ALL_STATS)}
    };
    // TODO: Other objects?
    return choose([
      gsp, gsp, gsp, gsp
    ])();
  }

  // == update logic ==
  #updateVisibilityAndPossibleMoves() {
    for (let x = 0; x < this.cells.size.w; x++) {
      for (let y = 0; y < this.cells.size.h; y++) {
        let position = new Point(x, y);
        let data = this.cells.get(position);

        data.nextMoveAccessible = false;
        if (
          Math.abs(x - this.player.x) <= 1 &&
          Math.abs(y - this.player.y) <= 1
        ) {
          data.revealed = true;
          if (!this.player.equals(position)) {
            data.nextMoveAccessible = true;
          }
        }

      }
    }
  }

  #collectResources() {
    let present = this.cells.get(this.player);
    if (present.content.type == "statPickup") {
      let stat = present.content.stat;
      let amount = 1;
      present.content = {type: "empty"};
      getPlayerProgress().spendBlood(90);
      getPlayerProgress().add(stat, amount);
    }
  }

  movePlayerTo(newPosition: Point) {
    this.player = newPosition;
    getPlayerProgress().spendBlood(10);
    this.#updateVisibilityAndPossibleMoves();
    this.#collectResources();
  }

  // draw
  update() {
    this.frame += 1;
    this.drawpile.clear();

    let globalOffset =
      new Point(this.player.x * MAP_CELL_ONSCREEN_SIZE.w, this.player.y * MAP_CELL_ONSCREEN_SIZE.h).offset(
        new Point(-192, -192)
      )

    let map = this.cells;
    for (let y = 0; y < map.size.h; y += 1) {
      for (let x = 0; x < map.size.w; x += 1) {
        let cellOffset = new Point(x * MAP_CELL_ONSCREEN_SIZE.w, y * MAP_CELL_ONSCREEN_SIZE.h).offset(globalOffset.negate());
        let cellData = this.cells.get(new Point(x, y))
        let belowIsBlock = true;
        if (y < map.size.h - 1) {
          let below = this.cells.get(new Point(x, y + 1));
          belowIsBlock = !below.revealed || below.content.type == "block";
        }

        this.#drawMapCell(cellOffset, new Point(x, y), cellData, belowIsBlock);
      }
    }
    this.#drawPlayer(globalOffset);

    this.drawpile.executeOnClick();
  }

  draw() {
    this.drawpile.draw()
  }

  #drawMapCell(
    cellOffset: Point,
    mapPosition: Point,
    cellData: MapCell,
    belowIsBlock: boolean
  ) {
    const OFFSET_FLOOR = -256;
    const OFFSET_AIR = 0;
    const depth = mapPosition.y;
    const onFloor = OFFSET_FLOOR + depth;
    const inAir = OFFSET_AIR + depth;

    if (!cellData.revealed) {
      return;
    }

    let cellTopLeft = cellOffset.offset(new Size(-MAP_CELL_ONSCREEN_SIZE.w / 2, -MAP_CELL_ONSCREEN_SIZE.h / 2));
    let cellSize = MAP_CELL_ONSCREEN_SIZE;

    if (cellData.content.type == "block") {
      if (!belowIsBlock) {
        this.drawpile.add(inAir, () => {
          D.drawSprite(sprDrips, cellOffset.offset(new Point(0, -cellSize.h)), 1, {xScale: 3, yScale: 3})
        })
      }
      return;
    }

    // draw inset zone
    this.drawpile.addClickable(onFloor,
      (hover: boolean) => {
        D.fillRect(cellTopLeft, cellSize, hover ? FG_TEXT : BG_INSET)
      },
      new Rect(cellTopLeft, cellSize),
      cellData.nextMoveAccessible,
      () => {
        this.movePlayerTo(mapPosition)
      }
    );

    if (belowIsBlock) {
      // draw the underhang
      this.drawpile.add(onFloor, () => {
        D.drawSprite(sprDrips, cellOffset.offset(new Point(0, cellSize.h/2)), 0, {xScale: 3, yScale: 3})
      })
    }

    if (cellData.content.type == "statPickup") {
      let content = cellData.content;
      let extraXOffset = 0; // Math.cos(this.frame / 80 + mapPosition.x + mapPosition.y) * 1;
      let extraYOffset = Math.sin(this.frame / 50 + mapPosition.x * 2+ mapPosition.y * 0.75) * 6 - 18;
      this.drawpile.add(inAir, () => {
        D.drawSprite(
          sprStatPickup,
          cellOffset.offset(new Point(extraXOffset, extraYOffset)),
          ALL_STATS.indexOf(content.stat),
          {
            xScale: 3,
            yScale: 3,
          }
        )
      });
    }
  }

  #drawPlayer(globalOffset: Point) {
    let cellOffset = new Point(
      this.player.x * MAP_CELL_ONSCREEN_SIZE.w,
      this.player.y * MAP_CELL_ONSCREEN_SIZE.h
    ).offset(globalOffset.negate())
    this.drawpile.add(this.player.y, () => {
      D.drawSprite(
        sprRaccoonWalking,
        cellOffset.offset(new Point(0, 22)),
        0, {
          xScale: 3,
          yScale: 3
        }
      )
    });
  }
}

const MAP_CELL_ONSCREEN_SIZE: Size = new Size(96, 48)

function choose<T>(array: Array<T>): T  {
  if (array.length == 0) {
    throw `array cannot have length 0 for choose`
  }
  return array[Math.floor(Math.random() * array.length)]
}
